# turtle.py 
# $Revision: 1.3 $
#
# LogoMation-like turtle graphics

import math as _math
import Sketch as _Sketch

class Error(Exception):
    pass

def _append_path(dest, src, start = 0, end = -1, copy_selection = 1):
    if start < 0:
        start = src.len + start
    if end < 0:
        end = src.len + end
    for i in range(start, end + 1):
        type, control, node, cont = src.Segment(i)
        dest.AppendSegment(type, control, node, cont)
        if copy_selection:
            dest.select_segment(-1, src.SegmentSelected(i))
            
def _copy_paths(src):
    newpaths = []
    for path in src:
        new = _Sketch.CreatePath()
        _append_path(new, path)
        newpaths.append(new)
    return tuple(newpaths)
            

class Pen:
    def __init__(self, doc_or_obj):
        if doc_or_obj.is_GraphicsObject:
            obj = doc_or_obj
            doc = obj.document
        else:
            obj = None
            doc = doc_or_obj
        self._document = doc
        self._object = obj
        self._functions = []
        self._origin = (0,0)
        self._fullcircle = 360.
        self.reset()
        self.degrees()

    def reset(self):
        "Resets turtle parameters to default values"
        self._start_new = 1
        self._angle = 0
        self._position = self._origin
        self._drawing = 1

    def clear(self):
        "Removes drawing"
        doc = self._document
        doc.RemoveSelected()

    def set_object(self, obj=None):
        self._object = obj
        
    def degrees(self, fullcircle=360.0):
        "Use degrees instead of radians"
        normal = self._angle / self._fullcircle
        self._fullcircle = fullcircle
        self._invradian = _math.pi / (fullcircle * 0.5)
        self._angle = normal * fullcircle

    def radians(self):
        "Use radians instead of degrees"
        self.degrees(2.0*_math.pi)

    def forward(self, distance):
        "Move turtle forward"
        x0, y0 = start = self._position
        x1 = x0 + distance * _math.cos(self._angle*self._invradian)
        y1 = y0 - distance * _math.sin(self._angle*self._invradian)
        self._goto(x1, y1)

    def backward(self, distance):
        "Move turtle backward"
        self.forward(-distance)

    def left(self, angle):
        "Turn turtle to the left"
        self._angle = (self._angle - angle) % self._fullcircle

    def right(self, angle):
        "Turn turtle to the right"
        self.left(-angle)

    def up(self):
        "Move pen up (no line is drawn when moving)"
        self._drawing = 0

    def down(self):
        "Move pen down (draw lines when moving)"
        self._drawing = 1
        self._start_new = 1

    def _new_obj(self, path):
        doc = self._document
        obj = _Sketch.PolyBezier(path)
        doc.BeginTransaction("turtle: insert")
        try:
            try:
                doc.Insert(obj)
                self._object = obj
            except:
                doc.AbortTransaction()    
        finally:
            doc.EndTransaction()
        self._start_new = 0

    def _get_paths(self, obj):
        paths = _copy_paths(obj.Paths())
        if self._start_new:
            path = _Sketch.CreatePath()
            paths = paths+(path,)
            self._start_new = 0
        else:
            path = paths[-1]
        return paths
        
    def _set_paths(self, paths, obj):
        doc = self._document
        doc.BeginTransaction("turtle: set path")
        try:
            try:
                doc.AddUndo(obj.SetPaths(paths))
            except:
                doc.AbortTransaction()    
        finally:
            doc.EndTransaction()

    def bezier(self, radius, angle, l1=0, l2=0):
        """
        bezier(self, radius, angle, l1=0, l2=0)

        Draw a bezier line defined by radius and angle.
        l1 and l0 give the control points.
        """
        x0, y0 = start = self._position
        f = self._function
        invrad = self._invradian
        
        angle = -angle # XXX
        if radius < 0 and angle < 0:
            radius = -radius
            angle = -angle
        elif radius < 0:
            angle = -angle
        
        elif angle < 0:
            radius = -radius
        
        end = self._angle + angle
        xc = x0 - radius * _math.sin(self._angle * invrad)
        yc = y0 - radius * _math.cos(self._angle * invrad)
        x1 = xc + radius * _math.sin(end * invrad)
        y1 = yc + radius * _math.cos(end * invrad)

        if self._drawing:
            x = x0 + l1 * _math.cos(self._angle * invrad)
            y = y0 - l1 * _math.sin(self._angle * invrad)
            handle0 = f((x,y))
            
            x = x1 - l2 * _math.cos(end * invrad)
            y = y1 + l2 * _math.sin(end * invrad)
            handle1 = f((x,y))

            p0 = f((x0, y0))
            p1 = f((x1, y1))
            obj = self._object
            if obj is None:
                path = _Sketch.CreatePath()
                path.AppendLine(p0)
                path.AppendBezier(handle0, handle1, p1)
                self._new_obj(path)
            else:
                paths = self._get_paths(obj)
                path = paths[-1]
                path.AppendBezier(handle0, handle1, p1)
                self._set_paths(paths, obj)
                
        self._position = map(float, (x1, y1))
        self._angle = end
        
    def circle(self, radius, extent = None):
        """
        circle(self, radius, extent = None)

        Draw a circle. Extend is by default 360.
        """
        quarter = self._fullcircle / 4.
        invrad = self._invradian

        turn = 0
        if extent == None:
            extent = self._fullcircle
            
        elif extent < 0:
            self._angle = self._angle-2*quarter
            extent = -extent # self._fullcircle + extent
            turn = 1
            
        segments = int(extent / quarter)
        rest = extent % quarter
        
        s = quarter * invrad * abs(radius)
        l = 0.55197 * abs(radius)
        for i in range(segments):
            self.bezier(radius, quarter, l, l)
            
        s = rest * invrad * radius
        l = 0.55197 * abs(radius) * rest / quarter 
        self.bezier(radius, rest, l, l)
                    
        if turn:
            self._angle = self._angle-2*quarter
            
    def goto(self, *args):
        """
        goto(x,y) or goto((x,y))
        
        Move turtle to x,y.
        """
        if len(args) == 1:
            try:
                x, y = args[0]
            except:
                raise Error, "bad point argument: %s" % `args[0]`
        else:
            try:
                x, y = args
            except:
                raise Error, "bad coordinates: %s" % `args[0]`
        x0, y0 = self._origin
        self._goto(x0+x, y0+y)

    def _function(self, x,y):
        for f in self._functions:
            x,y = f(x,y)
        return x,y

    def _goto(self, x1, y1):
        x0, y0 = start = self._position
        f = self._function
        if self._drawing:
            obj = self._object
            if obj is None:
                path = _Sketch.CreatePath()
                path.AppendLine(f(x0,y0))
                path.AppendLine(f(x1,y1))
                self._new_obj(path)
            else:
                paths = self._get_paths(obj)
                path = paths[-1]
                path.AppendLine(f(x0,y0))
                path.AppendLine(f(x1,y1))
                self._set_paths(paths, obj)
                
        self._position = map(float, (x1, y1))

    def goto_end(self):
        obj = self._object
        if obj is not None:
            path = self._get_paths(obj)[-1]
            i = path.len-1
            type, control, node, cont = path.Segment(i)
        self._position = node
 

    def angle(self, value=None):
        if value is not None:
            self._angle = value
        return self._angle

    def push_function(self, f):
        self._functions.append(f)
    def pop_function(self):
        return self._functions.pop()
    def set_function(self, f=None):
        if f is None:
            self._functions = []
        else:
            self._functions = [f]
        
    
            
                
_document = None
_pen = None

def create_pen(document):
    global _document, _pen
    _document = document
    _pen = Pen(_document)
    return _pen
    
def _getpen():
    global _pen, _document
    pen = _pen
    if not pen:
        if not _document:
            raise Error, 'You need to create a pen first. ' \
                         'Use "create_pen(doc)".'
        _pen = pen = Pen(_document)
    return pen

def degrees(): _getpen().degrees()
def radians(): _getpen().radians()
def reset(): _getpen().reset()
def clear(): _getpen().clear()
def forward(distance): _getpen().forward(distance)
def backward(distance): _getpen().backward(distance)
def left(angle): _getpen().left(angle)
def right(angle): _getpen().right(angle)
def up(): _getpen().up()
def down(): _getpen().down()
def circle(radius, extent=None): _getpen().circle(radius, extent)
def goto(*args): apply(_getpen().goto, args)
def bezier(*args): apply(_getpen().bezier, args)
def angle(*args): 
    return apply(_getpen().angle, args)
    
        
